#ifndef QNNTPGRAB_GLUE_H
#define QNNTPGRAB_GLUE_H

#include <QObject>
#include <QString>
#include <QDateTime>

class QNNTPGrabGluePrivate;

class QNNTPGrabGlue : public QObject
{
    Q_OBJECT

public:
    QNNTPGrabGlue();
    ~QNNTPGrabGlue();

    enum NNTPDisconnectType {
        DISCONNECT_NORMAL,
        DISCONNECT_NO_SUCH_HOST,
        DISCONNECT_CONNECTION_REFUSED,
        DISCONNECT_TOO_MANY_CONNECTIONS,
        DISCONNECT_CONNECT_TIMEOUT,
        DISCONNECT_READ_ERROR,
        DISCONNECT_READ_TIMEOUT,
        DISCONNECT_WRITE_ERROR,
        DISCONNECT_IDLE_TIMEOUT,
        DISCONNECT_INVALID_MSG,
        DISCONNECT_LOGIN_FAILURE,
        DISCONNECT_ERROR_SSL_INITIALISE,
        DISCONNECT_UNEXPECTED
    };

    enum TaskState {
        TASK_STATE_WAITING_FOR_DOWNLOAD,
        TASK_STATE_DOWNLOADING,
        TASK_STATE_WAITING_FOR_DECODE,
        TASK_STATE_DECODING,
        TASK_STATE_FINISHED_COMPLETE,
        TASK_STATE_FINISHED_INCOMPLETE,
        TASK_STATE_FINISHED_NO_PARTS_AVAIL,
        TASK_STATE_SKIPPED
    };

    enum LogLevel {
        LOG_LEVEL_ALL = 0,
        LOG_LEVEL_INFO,
        LOG_LEVEL_WARNING,
        LOG_LEVEL_ERROR,
        LOG_LEVEL_FATAL,
        LOG_LEVEL_DEBUG
    };

    enum SchedularState {
        SCHEDULAR_STATE_RUNNING,
        SCHEDULAR_STATE_STOPPING,
        SCHEDULAR_STATE_STOPPED
    };

    enum ServerPriority {
        SERVER_PRIORITY_HIGH,
        SERVER_PRIORITY_NORMAL,
        SERVER_PRIORITY_LOW
    };

    struct ConfigServer {
        QString servername;
        QString hostname;
        int port;
        QString username;
        QString password;
        int max_threads;
        ServerPriority priority;
        bool send_group_command;
        bool use_ssl;
        bool enabled;
    };

    struct ConfigOpts {
        QString download_directory;
        QString temp_directory;
        bool enable_intelligent_par2_downloading;
        bool enable_par2_repair;
        QString auto_import_directory;
        bool enable_auto_import;
        bool move_file_after_auto_import;
        bool enable_auto_unpack;
        bool enable_bandwidth_shaping;
        int max_bandwidth;
        bool enable_webserver;
        int webserver_port;
        bool enable_logger;
        bool auto_remove_files_after_repair;
        bool auto_remove_files_after_unpack;
        bool auto_remove_collections_after_download;
    };

    struct Folder {
        QString folder;
        bool has_subfolders;
    };

    struct File {
        QString collection_name;
        QString subject;
        QString poster;
        QDateTime stamp;
        quint64 file_size;
        quint64 file_size_remaining;
        int position;
        int num_parts_total;
        int num_parts_downloaded;
        int num_parts_failed;
        QNNTPGrabGlue::TaskState status;
        QString filename;
        QList<QString> groups;
    };

    struct Collection {
        QString collection_name;
        QString poster;
        quint64 total_size;
        quint64 total_size_remaining;
        int position;
        QList<QNNTPGrabGlue::File> files;
    };

    struct Part {
        QString message_id;
        int size;
        int part_num;
    };

    struct NZBFile {
        QString subject;
        QString poster;
        QDateTime stamp;
        quint64 file_size;
        QList<QString> groups;
        QList<QNNTPGrabGlue::Part> segments;
    };

    struct NZB {
        QList<QNNTPGrabGlue::NZBFile> files;
    };

    struct PluginInfo {
        QString name;
        QString version;
        QString author;
        QString url;
        QString description;

        bool is_loaded;
        bool is_persistent;
    };

    bool init(QString hostname, int port, QString username, QString password, bool use_ssl, QString &errmsg, QString &warnings);
    bool getIsConnected();
    bool getIsStandalone();
    void killNNTPGrabServer();

    QList<QString> configGetAvailServers();
    bool configGetServerInfo(QString servername, QNNTPGrabGlue::ConfigServer &server);
    bool configAddServer(QNNTPGrabGlue::ConfigServer new_server, QString &errmsg);
    bool configDelServer(QString servername, QString &errmsg);
    bool configEditServer(QString servername, QNNTPGrabGlue::ConfigServer server, QString &errmsg);

    QNNTPGrabGlue::ConfigOpts configGetOpts();
    void configSetOpts(QNNTPGrabGlue::ConfigOpts opts);

    bool configGetFolderListing(QString parent, QList<QNNTPGrabGlue::Folder> folders);

    bool startSchedular();
    bool stopSchedular(bool wait);
    SchedularState getSchedularState();

    bool schedularAddTaskToQueue(QString collection_name, QString subject, QString poster, QDateTime stamp, quint64 file_size, QList<QString> groups, QList<QNNTPGrabGlue::Part> parts, QString &errmsg);
    bool schedularDelTaskFromQueue(QString collection_name, QString subject, QString &errmsg);
    bool schedularRestartTask(QString collection_name, QString subject, QString &errmsg);
    bool schedularMarkTaskOptional(QString collection_name, QString subject, bool is_optional);
    bool schedularSaveQueue(QString &errmsg);

    QList<QNNTPGrabGlue::Collection> schedularGetAllTasks(void);

    bool schedularMoveTask(QString collection_name_src, QString subject_src, QString collection_name_dest, int position_dest);
    bool schedularMoveCollection(QString collection_name, int new_position);

    QList<QString> getAvailPlugins();
    bool getPluginInfo(QString plugin_name, QNNTPGrabGlue::PluginInfo &plugin_info);
    bool loadPlugin(QString plugin_name, QString &errmsg);
    bool unloadPlugin(QString plugin_name, QString &errmsg);
    bool setPluginPersistent(QString plugin_name, bool persistent);

    void setEmitLogMessages(bool val);

    static QString utilsCalculateFileSize(quint64 file_size);
    static int utilsCalculateEstimatedTimeRemaining(int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, quint64 file_size);
    static QString utilsGetReadableTimeRemaining(int estimated_time_remaining);
    static QString utilsGetReadableFinishedTime(int estimated_time_remaining);

    static bool utilsParseNZBFile(QString nzb_data, QNNTPGrabGlue::NZB &nzbfile, QString &errmsg);
    static QString utilsSanitizeCollectionName(QString collection_name);

    static bool utilsTestIsServerAlreadyRunning(void);

    static bool utilsPerformShutdown(void);

signals:
    void connectionLost(QString errmsg);

    void configChanged();
    void allDownloadsCompleted();
    void partDownloadStart(QString servername, int conn_id, QString collection_name, QString subject, int part_num);
    void partDone(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int size);
    void partFailed(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int size, bool all_servers_tried);
    void partProgressUpdate(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int bytes_downloaded, int bytes_total);
    void trafficMonitorUpdate(int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, QDateTime stamp, double average);
    void collectionAdded(QString collection_name, QString poster);
    void collectionModified(QString collection_name, QString poster);
    void collectionRemoved(QString collection_name);
    void fileAdded(QString collection_name, QString subject, QString poster, QDateTime stamp, quint64 file_size, quint64 total_size, quint64 total_size_remaining, QNNTPGrabGlue::TaskState state, int num_parts, QList<QString> groups);
    void fileRemoved(QString collection_name, QString subject, quint64 total_size, quint64 total_size_remaining);
    void fileDownloadStateUpdate(QString collection_name, QString subject, int num_parts_total, int num_parts_done, int num_parts_failed, quint64 file_size, quint64 file_size_remaining, quint64 total_size, quint64 total_size_remaining);
    void fileStateChanged(QString collection_name, QString subject, QString real_filename, QNNTPGrabGlue::TaskState old_state, QNNTPGrabGlue::TaskState new_state, quint64 file_size_remaining, quint64 total_size, quint64 total_size_remaining);
    void connectionConnecting(QString servername, int conn_id);
    void connectionConnected(QString servername, int conn_id, QString welcome_msg);
    void connectionDisconnect(QString servername, int conn_id, QNNTPGrabGlue::NNTPDisconnectType disconnect_type, QString reason);
    void schedularStateChanged(QNNTPGrabGlue::SchedularState new_state, QString reason);
    void logMessage(QString component, QNNTPGrabGlue::LogLevel log_level, QString msg);
    void fileMoved(QString orig_collection_name, QString subject, QString new_collection_name, int old_position, int new_position);
    void collectionMoved(QString collection_name, int old_position, int new_position);
    void collectionDownloaded(QString collection_name);
    void pluginLoaded(QString plugin_name, bool is_persistent);
    void pluginUnloaded(QString plugin_name);
    void pluginEvent(QString plugin_name, QString event_name, QList<QString> values);

private:
    Q_DECLARE_PRIVATE(QNNTPGrabGlue)
    QNNTPGrabGluePrivate * const d_ptr;
};

#endif // QNNTPGRAB_GLUE_H
