#include <time.h>

#include "WidgetConnections.h"
#include "ui_WidgetConnections.h"
#include "ConfigGuiOpts.h"

WidgetConnections::WidgetConnections(QNNTPGrabGlue *glue, QWidget *parent) :
        QWidget(parent),
        ui(new Ui::WidgetConnections)
{
    this->glue = glue;
    ui->setupUi(this);
    ui->treeWidget->hideColumn(4);

    progressbarDelegate = new ProgressBarDelegate(this);
    ui->treeWidget->setItemDelegateForColumn(2, progressbarDelegate);

    connect(glue, SIGNAL(configChanged()), SLOT(onConfigChanged()), Qt::QueuedConnection);
    connect(glue, SIGNAL(connectionConnecting(QString,int)), SLOT(onConnectionConnecting(QString,int)), Qt::QueuedConnection);
    connect(glue, SIGNAL(connectionConnected(QString,int,QString)), SLOT(onConnectionConnected(QString,int,QString)), Qt::QueuedConnection);
    connect(glue, SIGNAL(connectionDisconnect(QString,int,QNNTPGrabGlue::NNTPDisconnectType,QString)), SLOT(onConnectionDisconnect(QString,int,QNNTPGrabGlue::NNTPDisconnectType,QString)), Qt::QueuedConnection);
    connect(glue, SIGNAL(partDownloadStart(QString,int,QString,QString,int)), SLOT(onPartDownloadStart(QString,int,QString,QString,int)), Qt::QueuedConnection);
    connect(glue, SIGNAL(partProgressUpdate(QString,int,QString,QString,int,int,int)), SLOT(onPartProgressUpdate(QString,int,QString,QString,int,int,int)), Qt::QueuedConnection);
    connect(glue, SIGNAL(partDone(QString,int,QString,QString,int,int)), SLOT(onPartDone(QString,int,QString,QString,int,int)), Qt::QueuedConnection);
    connect(glue, SIGNAL(partFailed(QString,int,QString,QString,int,int,bool)), SLOT(onPartFailed(QString,int,QString,QString,int,int,bool)), Qt::QueuedConnection);

    reloadGuiOpts();

    num_disconnected_connections = 0;
}

WidgetConnections::~WidgetConnections()
{
    delete ui;
}


void WidgetConnections::onConfigChanged()
{
    QList<QString> servers = glue->configGetAvailServers();

    ui->treeWidget->clear();

    for (int i = 0; i < servers.length(); i++) {
        QNNTPGrabGlue::ConfigServer server;

        if (!glue->configGetServerInfo(servers.at(i), server)) {
            /* Shouldn't happen */
            qDebug("Unable to find server with name '%s'\n", servers.at(i).toUtf8().data());
            continue;
        }

        if (!server.enabled) {
            continue;
        }

        for (int j = 0; j < server.max_threads; j++) {
            QTreeWidgetItem *item;

            item = new QTreeWidgetItem();
            item->setText(0, server.servername);
            item->setText(1, tr("Unused"));
            item->setText(2, "0");
            item->setText(3, "");
            item->setText(4, "0");

            ui->treeWidget->addTopLevelItem(item);

            item->setHidden(hideInactiveConnections);
        }
    }
}

void WidgetConnections::onConnectionConnecting(QString servername, int conn_id)
{
    Q_UNUSED(servername);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(1, tr("Connecting"));
}

void WidgetConnections::onConnectionConnected(QString servername, int conn_id, QString welcome_msg)
{
    Q_UNUSED(servername);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(1, tr("Connected"));
    item->setText(3, welcome_msg);
}

void WidgetConnections::onConnectionDisconnect(QString servername, int conn_id, QNNTPGrabGlue::NNTPDisconnectType disconnect_type, QString reason)
{
    Q_UNUSED(servername);
    Q_UNUSED(disconnect_type);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setText(1, tr("Disconnected"));
    item->setText(3, reason);
    item->setText(4, QString("%1").arg(time(NULL)));

    if (num_disconnected_connections == 0) {
        startTimer(1000);
    }
    num_disconnected_connections++;
}

void WidgetConnections::onPartDownloadStart(QString servername, int conn_id, QString collection_name, QString subject ,int part_num)
{
    Q_UNUSED(servername);
    Q_UNUSED(collection_name);
    Q_UNUSED(part_num);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(1, tr("Downloading"));
    item->setText(2, "0");
    item->setText(3, subject);
}

void WidgetConnections::onPartProgressUpdate(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int bytes_downloaded, int bytes_total)
{
    double progress;
    int progress_int;

    Q_UNUSED(servername);
    Q_UNUSED(collection_name);
    Q_UNUSED(subject);
    Q_UNUSED(part_num);

    progress = bytes_downloaded;
    progress /= bytes_total;
    progress *= 100;
    progress_int = (int) progress;

    if (progress_int > 100) {
        qDebug(__FILE__ ":%i progress overflow!\n", __LINE__);
        progress_int = 100;
    }

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(2, QString("%1").arg(progress_int));
}

void WidgetConnections::onPartDone(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int size)
{
    Q_UNUSED(servername);
    Q_UNUSED(collection_name);
    Q_UNUSED(subject);
    Q_UNUSED(part_num);
    Q_UNUSED(size);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(1, tr("Idle"));
    item->setText(2, "0");
    item->setText(3, "");
}

void WidgetConnections::onPartFailed(QString servername, int conn_id, QString collection_name, QString subject, int part_num, int size, bool all_servers_tried)
{
    Q_UNUSED(servername);
    Q_UNUSED(collection_name);
    Q_UNUSED(subject);
    Q_UNUSED(part_num);
    Q_UNUSED(size);
    Q_UNUSED(all_servers_tried);

    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(conn_id - 1);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    item->setHidden(false);
    item->setText(1, tr("Idle"));
    item->setText(2, "0");
    item->setText(3, "");
}

void WidgetConnections::reloadGuiOpts()
{
    ConfigGuiOpts opts;
    hideInactiveConnections = opts.getHideInactiveConnections();
    onConfigChanged();
}

void WidgetConnections::timerEvent(QTimerEvent *event)
{
    /* Check every second if we need to mark a disconnected connection as 'unused' */
    time_t now = time(NULL);
    QTreeWidgetItem *item = ui->treeWidget->topLevelItem(0);

    if (!item) {
        /* A configuration change probably just occured causing fewer connections slots to be available than before */
        return;
    }

    do {
        time_t stamp = item->text(4).toInt();

        if (stamp > 0 && stamp + 10 <= now) {
            if (hideInactiveConnections) {
                item->setHidden(true);
            }

            item->setText(1, tr("Unused"));
            item->setText(2, "0");
            item->setText(3, "");
            item->setText(4, "0");

            num_disconnected_connections--;
            Q_ASSERT(num_disconnected_connections >= 0);
            if (num_disconnected_connections == 0) {
                killTimer(event->timerId());
            }
        }
    } while ((item = ui->treeWidget->itemBelow(item)));
}

void WidgetConnections::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

