#include "WidgetNZBCreator.h"
#include "ui_WidgetNZBCreator.h"
#include "DialogImportNZB.h"
#include <QFuture>
#include <QtConcurrentRun>
#include <QFutureWatcher>

enum {
    FIELD_IMPORT,
    FIELD_PERCENTAGE_COMPLETE,
    FIELD_SUBJECT,
    FIELD_POSTER,
    FIELD_NEWSGROUP,
    FIELD_FILE_SIZE,
    FIELD_STAMP,
    FIELD_FILE_ID,
    FIELD_COLLECTION_NAME_SHORT,
    FIELD_IS_SINGLE_FILE
};

WidgetNZBCreator::WidgetNZBCreator(QNNTPGrabGlue *glue, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WidgetNZBCreator)
{
    ui->setupUi(this);

    this->glue = glue;
    this->results_per_page = 0;

    connect(ui->entryQuery, SIGNAL(textChanged(QString)), SLOT(entryQuery_textChanged(QString)));
    connect(ui->treeWidget, SIGNAL(itemChanged(QTreeWidgetItem*,int)), SLOT(treeWidget_itemChanged(QTreeWidgetItem*,int)));
    connect(ui->treeWidget, SIGNAL(expanded(QModelIndex)), SLOT(treeWidget_expanded(QModelIndex)));
    connect(ui->btnSearch, SIGNAL(clicked()), SLOT(btnSearch_clicked()));
    connect(ui->btnImport, SIGNAL(clicked()), SLOT(btnImport_clicked()));
    connect(ui->btnNewerResults, SIGNAL(clicked()), SLOT(btnNewerResults_clicked()));
    connect(ui->btnOlderResults, SIGNAL(clicked()), SLOT(btnOlderResults_clicked()));
    connect(ui->btnSelectAll, SIGNAL(clicked()), SLOT(btnSelectAll_clicked()));
    connect(ui->btnSelectNone, SIGNAL(clicked()), SLOT(btnSelectNone_clicked()));

    progressbarDelegate = new ProgressBarDelegate(this);
    ui->treeWidget->setItemDelegateForColumn(FIELD_PERCENTAGE_COMPLETE, progressbarDelegate);

    ui->treeWidget->setColumnHidden(FIELD_FILE_ID, true);
    ui->treeWidget->setColumnHidden(FIELD_COLLECTION_NAME_SHORT, true);
    ui->treeWidget->setColumnHidden(FIELD_IS_SINGLE_FILE, true);

    /* Manually set the size of various columns */
    ui->treeWidget->setColumnWidth(FIELD_IMPORT, 70);
    ui->treeWidget->resizeColumnToContents(FIELD_PERCENTAGE_COMPLETE);
    ui->treeWidget->setColumnWidth(FIELD_NEWSGROUP, 150);
    ui->treeWidget->resizeColumnToContents(FIELD_FILE_SIZE);

    connect(&watcher_init, SIGNAL(finished()), SLOT(performInit_finished()));
    connect(&watcher_search, SIGNAL(finished()), SLOT(performSearch_finished()));
    connect(&watcher_import, SIGNAL(finished()), SLOT(performImport_finished()));

    pool_details.setMaxThreadCount(5);

    future_init = QtConcurrent::run(this, &WidgetNZBCreator::performInit);
    watcher_init.setFuture(future_init);
    setEnabled(false);
}

WidgetNZBCreator::~WidgetNZBCreator()
{
    delete ui;
}

void WidgetNZBCreator::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

WidgetNZBCreator::InitResult WidgetNZBCreator::performInit()
{
    WidgetNZBCreator::InitResult ret;

    /* Make sure that the API version is okay first */
    if (!(ret.ret = nzbcreator.checkApiVersion(ret.api_okay, ret.errmsg))) {
        return ret;
    }
    if (!ret.api_okay) {
        return ret;
    }

    /* Retrieve maintenance information */
    ret.ret = nzbcreator.getMaintenanceInfo(ret.errmsg, ret.maintenance_info);

    return ret;
}

void WidgetNZBCreator::performInit_finished()
{
    WidgetNZBCreator::InitResult result = future_init.result();
    QString msg;

    if (!result.ret) {
        msg = tr("Initialisation failed: ") + result.errmsg;
        return;
    }

    if (!result.api_okay) {
        msg = tr("A newer version of NNTPGrab is required to use the Online Search feature");
    }

    if (result.maintenance_info.down_for_maintenance) {
        msg = result.maintenance_info.reason;
    }

    if (!msg.isEmpty()) {
        QTreeWidgetItem *collection_item = new QTreeWidgetItem();

        collection_item->setText(FIELD_SUBJECT, msg);

        ui->treeWidget->addTopLevelItem(collection_item);

        return;
    }

    setEnabled(true);
}

void WidgetNZBCreator::btnSearch_clicked()
{
    QNZBCreator::SearchOpts opts;

    opts.query = ui->entryQuery->text();
    opts.username = "guest";
    opts.password = "guest";
    opts.search_only_in = QNZBCreator::SEARCH_ONLY_IN_FILENAMES;
    opts.max_age = 14;
    opts.minimal_file_size = 0;
    opts.maximal_file_size = 0;
    opts.percentage_complete = 0;
    opts.group_to_search = 0;
    opts.file_type = QNZBCreator::FILE_TYPE_UNKNOWN;
    opts.offset = 0;

    if (opts.query.length() < 4) {
        qWarning("%s", tr("The search query must be at least 4 characters long").toUtf8().data());
        return;
    }

    ui->treeWidget->clear();
    ui->btnSearch->setEnabled(false);
    ui->btnNewerResults->setEnabled(false);
    ui->btnOlderResults->setEnabled(false);
    ui->btnSelectAll->setEnabled(false);
    ui->btnSelectNone->setEnabled(false);
    ui->btnImport->setEnabled(false);

    cached_search_opts = opts;

    future_search = QtConcurrent::run(this, &WidgetNZBCreator::performSearch, opts);
    watcher_search.setFuture(future_search);
}

WidgetNZBCreator::SearchResult WidgetNZBCreator::performSearch(QNZBCreator::SearchOpts opts)
{
    WidgetNZBCreator::SearchResult ret;

    /* Make sure that the API version is okay first */
    bool api_okay;
    if (!nzbcreator.checkApiVersion(api_okay, ret.errmsg)) {
        ret.ret = false;
        return ret;
    }
    if (!api_okay) {
        ret.ret = false;
        ret.errmsg = tr("A newer version of NNTPGrab is required to use the Online Search feature");
        return ret;
    }

    /* Perform the real search */
    ret.ret = nzbcreator.performSearch(opts, ret.errmsg, ret.result);
    return ret;
}

void WidgetNZBCreator::performSearch_finished()
{
    WidgetNZBCreator::SearchResult result = future_search.result();

    if (!result.ret) {
        qWarning("%s", (tr("Performing Online Search failed: ") + result.errmsg).toUtf8().data());
        return;
    }

    if (result.result.number_of_results == 0) {
        qWarning("%s", tr("Search query returned no results").toUtf8().data());
        return;
    }

    ui->btnSelectAll->setEnabled(true);
    ui->btnSelectNone->setEnabled(true);
    ui->btnImport->setEnabled(true);

    results_per_page = result.result.results_per_page;

    if (result.result.offset > 0) {
        ui->btnNewerResults->setEnabled(true);
    }

    if (result.result.number_of_results > results_per_page &&
        result.result.offset + results_per_page < result.result.number_of_results) {

        ui->btnOlderResults->setEnabled(true);
    }

    Q_FOREACH(QNZBCreator::Collection collection, result.result.collections) {
        QTreeWidgetItem *collection_item = new QTreeWidgetItem();

        collection_item->setCheckState(FIELD_IMPORT, Qt::Unchecked);
        collection_item->setText(FIELD_SUBJECT, collection.collection_name);
        collection_item->setText(FIELD_COLLECTION_NAME_SHORT, collection.short_collection_name);
        collection_item->setText(FIELD_POSTER, collection.poster);
        collection_item->setText(FIELD_NEWSGROUP, collection.newsgroup);
        collection_item->setText(FIELD_FILE_SIZE, QNNTPGrabGlue::utilsCalculateFileSize(collection.total_size));
        collection_item->setText(FIELD_STAMP, collection.stamp.toString());
        collection_item->setText(FIELD_PERCENTAGE_COMPLETE, QString("%1").arg(collection.complete_percentage));
        collection_item->setText(FIELD_FILE_ID, QString("%1").arg(collection.collection_id));

        /* Don't show an expander for collections with only one file */
        if (collection.first_file_id != -1) {
            collection_item->setText(FIELD_FILE_ID, QString("%1").arg(collection.first_file_id));
            collection_item->setText(FIELD_IS_SINGLE_FILE, "1");
        } else {
            QTreeWidgetItem *file_item = new QTreeWidgetItem(collection_item);
            file_item->setText(FIELD_SUBJECT, "unknown");
        }

        ui->treeWidget->addTopLevelItem(collection_item);
    }

    ui->treeWidget->resizeColumnToContents(FIELD_SUBJECT);
    ui->treeWidget->resizeColumnToContents(FIELD_COLLECTION_NAME_SHORT);
    ui->treeWidget->resizeColumnToContents(FIELD_POSTER);
    ui->treeWidget->resizeColumnToContents(FIELD_FILE_SIZE);
    ui->treeWidget->resizeColumnToContents(FIELD_STAMP);
}

void WidgetNZBCreator::performGetDetails(QTreeWidgetItem *item, int collection_id, QDateTime stamp)
{
    WidgetNZBCreatorDetails *watcher = new WidgetNZBCreatorDetails(item, collection_id, stamp);

    item->child(0)->setText(FIELD_SUBJECT, tr("No information available yet. Please wait a moment..."));

    connect(watcher, SIGNAL(finished(bool,QTreeWidgetItem*,QList<QNZBCreator::File>,QString)), SLOT(performGetDetails_finished(bool,QTreeWidgetItem*,QList<QNZBCreator::File>,QString)), Qt::QueuedConnection);
    pool_details.start(watcher);
}

void WidgetNZBCreator::performGetDetails_finished(bool ret, QTreeWidgetItem *item, QList<QNZBCreator::File> files, QString errmsg)
{
    if (!ret) {
        qWarning("%s", (tr("Retrieving online search collection details failed: ") + errmsg).toUtf8().data());
        return;
    }

     Q_FOREACH(QNZBCreator::File file, files) {
         QTreeWidgetItem *file_item = new QTreeWidgetItem(item);

         /* Automatically select all the files if the placeholder is checked already */
         file_item->setCheckState(FIELD_IMPORT, item->checkState(FIELD_IMPORT));

         file_item->setText(FIELD_PERCENTAGE_COMPLETE, QString("%1").arg(file.complete_percentage));
         file_item->setText(FIELD_SUBJECT, file.subject);
         file_item->setText(FIELD_POSTER, file.poster);
         file_item->setText(FIELD_NEWSGROUP, item->text(FIELD_NEWSGROUP));
         file_item->setText(FIELD_FILE_SIZE, QNNTPGrabGlue::utilsCalculateFileSize(file.file_size));
         file_item->setText(FIELD_STAMP, file.stamp.toString());
         file_item->setText(FIELD_FILE_ID, QString("%1").arg(file.file_id));
     }

     /* Remove the placeholder item */
     item->removeChild(item->child(0));
}

void WidgetNZBCreator::btnImport_clicked()
{
    bool multiple_collections = FALSE;
    QList<int> file_ids;
    QString collection_name = "";
    QDateTime oldest_stamp;
    QDateTime newest_stamp;
    QDateTime stamp;

    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        QTreeWidgetItem *collection_item = ui->treeWidget->topLevelItem(i);

        /* Does this collection have any children? */
        if (collection_item->childCount() == 0) {
            if (collection_item->checkState(FIELD_IMPORT) == Qt::Checked) {
                file_ids.append(collection_item->text(FIELD_FILE_ID).toInt());

                if (collection_name.isEmpty()) {
                    collection_name = collection_item->text(FIELD_COLLECTION_NAME_SHORT);
                } else if (collection_name != collection_item->text(FIELD_COLLECTION_NAME_SHORT)) {
                    multiple_collections = TRUE;
                }

                stamp = QDateTime::fromString(collection_item->text(FIELD_STAMP));
                if (stamp.isValid() && (oldest_stamp.isNull() || stamp < oldest_stamp)) {
                    oldest_stamp = stamp;
                }

                if (stamp.isValid() && stamp > newest_stamp) {
                    newest_stamp = stamp;
                }
            }

            continue;
        }

        for (int j = 0; j < collection_item->childCount(); j++) {
            QTreeWidgetItem *file_item = collection_item->child(j);

            if (file_item->checkState(FIELD_IMPORT) == Qt::Checked) {
                file_ids.append(file_item->text(FIELD_FILE_ID).toInt());

                if (collection_name.isEmpty()) {
                    collection_name = collection_item->text(FIELD_COLLECTION_NAME_SHORT);
                } else if (collection_name != collection_item->text(FIELD_COLLECTION_NAME_SHORT)) {
                    multiple_collections = TRUE;
                }

                stamp = QDateTime::fromString(file_item->text(FIELD_STAMP));
                if (stamp.isValid() && (oldest_stamp.isNull() || stamp < oldest_stamp)) {
                    oldest_stamp = stamp;
                }

                if (stamp.isValid() && stamp > newest_stamp) {
                    newest_stamp = stamp;
                }
            }
        }

        /* Is this a single-file collection? */
        if (collection_item->text(FIELD_IS_SINGLE_FILE) != "" &&
            collection_item->checkState(FIELD_IMPORT) == Qt::Checked) {

            file_ids.append(collection_item->text(FIELD_FILE_ID).toInt());

            if (collection_name.isEmpty()) {
                collection_name = collection_item->text(FIELD_COLLECTION_NAME_SHORT);
            } else if (collection_name != collection_item->text(FIELD_COLLECTION_NAME_SHORT)) {
                multiple_collections = TRUE;
            }
        }
    }

    if (collection_name.isEmpty()) {
        qWarning("%s", tr("Please select one or more files which you want to import").toUtf8().data());
        return;
    }

    if (multiple_collections) {
        collection_name = ui->entryQuery->text();
    }

    ui->btnImport->setEnabled(false);
    ui->treeWidget->setEnabled(false);
    ui->btnNewerResults->setEnabled(false);
    ui->btnOlderResults->setEnabled(false);
    ui->btnSelectAll->setEnabled(false);
    ui->btnSelectNone->setEnabled(false);
    future_import = QtConcurrent::run(this, &WidgetNZBCreator::performImport, collection_name, file_ids, oldest_stamp, newest_stamp);
    watcher_import.setFuture(future_import);
}

WidgetNZBCreator::ImportResult WidgetNZBCreator::performImport(QString collection_name, QList<int> file_ids, QDateTime oldest_stamp, QDateTime newest_stamp)
{
    WidgetNZBCreator::ImportResult res;
    res.collection_name = collection_name;
    res.oldest_stamp = oldest_stamp;
    res.newest_stamp = newest_stamp;
    res.nzb_data = nzbcreator.generateNZB(file_ids, res.oldest_stamp, res.newest_stamp, res.errmsg);
    return res;
}

void WidgetNZBCreator::performImport_finished()
{
    WidgetNZBCreator::ImportResult res = future_import.result();

    if (res.nzb_data.isEmpty()) {
        qWarning("%s", (tr("Unable to import files: ") + res.errmsg).toUtf8().data());
        return;
    }

    DialogImportNZB d(glue);
    d.quickImport(res.collection_name, res.nzb_data, true);

    ui->btnImport->setEnabled(true);
    ui->treeWidget->setEnabled(true);
    if (cached_search_opts.offset > 0) {
        ui->btnNewerResults->setEnabled(true);
    }
    ui->btnOlderResults->setEnabled(true);
    ui->btnSelectAll->setEnabled(true);
    ui->btnSelectNone->setEnabled(true);
}

void WidgetNZBCreator::btnOlderResults_clicked()
{
    Q_ASSERT(results_per_page > 0);
    Q_ASSERT(cached_search_opts.offset >= 0);

    ui->treeWidget->clear();
    ui->btnSearch->setEnabled(false);
    ui->btnNewerResults->setEnabled(false);
    ui->btnOlderResults->setEnabled(false);
    ui->btnSelectAll->setEnabled(false);
    ui->btnSelectNone->setEnabled(false);
    ui->btnImport->setEnabled(false);

    cached_search_opts.offset += results_per_page;

    future_search = QtConcurrent::run(this, &WidgetNZBCreator::performSearch, cached_search_opts);
    watcher_search.setFuture(future_search);
}

void WidgetNZBCreator::btnNewerResults_clicked()
{
    Q_ASSERT(results_per_page > 0);
    Q_ASSERT(cached_search_opts.offset >= results_per_page);

    ui->treeWidget->clear();
    ui->btnSearch->setEnabled(false);
    ui->btnNewerResults->setEnabled(false);
    ui->btnOlderResults->setEnabled(false);
    ui->btnSelectAll->setEnabled(false);
    ui->btnSelectNone->setEnabled(false);
    ui->btnImport->setEnabled(false);

    cached_search_opts.offset -= results_per_page;

    future_search = QtConcurrent::run(this, &WidgetNZBCreator::performSearch, cached_search_opts);
    watcher_search.setFuture(future_search);
}

void WidgetNZBCreator::btnSelectAll_clicked()
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        ui->treeWidget->topLevelItem(i)->setCheckState(FIELD_IMPORT, Qt::Checked);
    }
}

void WidgetNZBCreator::btnSelectNone_clicked()
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        ui->treeWidget->topLevelItem(i)->setCheckState(FIELD_IMPORT, Qt::Unchecked);
    }
}

void WidgetNZBCreator::entryQuery_textChanged(QString text)
{
    if (text.isEmpty()) {
        ui->btnSearch->setEnabled(false);
    } else {
        ui->btnSearch->setEnabled(true);
    }
}

void WidgetNZBCreator::treeWidget_itemChanged(QTreeWidgetItem *item, int column)
{
    /* Automatically check files when the collection item is checked */
    if (column != FIELD_IMPORT || item->childCount() == 0) {
        return;
    }

    for (int i = 0; i < item->childCount(); i++) {
        QTreeWidgetItem *item_child = item->child(i);
        item_child->setCheckState(FIELD_IMPORT, item->checkState(column));

        /* Do we know the contents of this collection yet? */
        if (i == 0 && item_child->text(FIELD_SUBJECT) == "unknown") {
            /* Start a job to retrieve the details */
            qint64 collection_id = item->text(FIELD_FILE_ID).toInt();
            QDateTime stamp = QDateTime::fromString(item->text(FIELD_STAMP));
            performGetDetails(item, collection_id, stamp);
        }
    }
}

void WidgetNZBCreator::treeWidget_expanded(QModelIndex idx)
{
    /* Check if the child already has some real data */
    QModelIndex idx_child = idx.child(0, 0);
    QTreeWidgetItem *item_child = ui->treeWidget->getItemByIndex(idx_child);
    Q_ASSERT(item_child != NULL);

    if (item_child->text(FIELD_SUBJECT) != "unknown") {
        /* Nothing to do as the real data is already there */
        return;
    }

    /* Get the collection_id from the parent */
    QTreeWidgetItem *item_parent = ui->treeWidget->getItemByIndex(idx);
    Q_ASSERT(item_parent != NULL);

    qint64 collection_id = item_parent->text(FIELD_FILE_ID).toInt();
    QDateTime stamp = QDateTime::fromString(item_parent->text(FIELD_STAMP));
    performGetDetails(item_parent, collection_id, stamp);
}
