#include <QtCore>
#include <QNetworkAccessManager>
#include <QUrl>
#include <QNetworkRequest>
#include <QXmlStreamReader>
#include <time.h>

#include "WidgetUpdateInfo.h"
#include "ui_WidgetUpdateInfo.h"
#include "config.h"
#include "version.h"

WidgetUpdateInfo::WidgetUpdateInfo(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WidgetUpdateInfo)
{
    ui->setupUi(this);

    /* Perform an update check */
    QNetworkAccessManager* nam;
    nam = new QNetworkAccessManager(this);
    QObject::connect(nam, SIGNAL(finished(QNetworkReply*)),SLOT(finishedSlot(QNetworkReply*)));

    QUrl url("http://www.nntpgrab.nl/info.php");
    nam->get(QNetworkRequest(url));
    //nam->deleteLater();
}

WidgetUpdateInfo::~WidgetUpdateInfo()
{
    delete ui;
}

void WidgetUpdateInfo::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void WidgetUpdateInfo::finishedSlot(QNetworkReply* reply)
{
    // Reading attributes of the reply
    // e.g. the HTTP status code
    QVariant statusCodeV = reply->attribute(QNetworkRequest::HttpStatusCodeAttribute);
    // Or the target URL if it was a redirect:
    QVariant redirectionTargetUrl = reply->attribute(QNetworkRequest::RedirectionTargetAttribute);
    // see CS001432 on how to handle this

    // no error received?
    if (reply->error() == QNetworkReply::NoError) {
        QByteArray bytes = reply->readAll();  // bytes
        processAndShowUpdateInfo(QString(bytes));
    } else {
        // handle errors here
        qDebug("Error occured while retrieving update information");
    }

    // We receive ownership of the reply object
    // and therefore need to handle deletion.
    reply->deleteLater();
}

class ChangeLog
{
public:
    int major_version;
    int minor_version;
    int micro_version;
    QList<QString> entries;
};

class UpdateInfo
{
public:
    ~UpdateInfo();

    int stable_major_version;
    int stable_minor_version;
    int stable_micro_version;
    time_t date_latest_stable_version;
    char date_latest_stable_version_str[128];

    int unstable_major_version;
    int unstable_minor_version;
    int unstable_micro_version;
    time_t date_latest_unstable_version;
    char date_latest_unstable_version_str[128];

    QList<ChangeLog*> changelogs;
};

UpdateInfo::~UpdateInfo()
{
    for (int i = 0; i < changelogs.count(); i++) {
        delete changelogs.at(i);
    }
}

bool WidgetUpdateInfo::processUpdateInfo(QString xml_data, UpdateInfo *info)
{
    QXmlStreamReader xml;
    QString currentTag;
    struct tm *now;
    ChangeLog *current_changelog = NULL;
    bool translated_entry_seen = FALSE;
    bool process_this_translated_entry = FALSE;
    QString lang_to_use = QLocale::system().name().left(2);

    xml.addData(xml_data);

    while (!xml.atEnd()) {
        xml.readNext();
        if (xml.isStartElement()) {
            if (xml.name() == "Version") {
                current_changelog = new ChangeLog();
                translated_entry_seen = FALSE;
            } else if (xml.name() == "Entry-i18n") {
                if (lang_to_use != xml.attributes().value("xml:lang").toString()) {
                    /* Different language, ignore */
                    continue;
                }

                translated_entry_seen = TRUE;
                process_this_translated_entry = TRUE;
            }

            currentTag = xml.name().toString();
        } else if (xml.isEndElement()) {
            if (xml.name() == "Version") {
                info->changelogs.append(current_changelog);
                current_changelog = NULL;
            }

            currentTag = "";
        } else if (xml.isCharacters() && !xml.isWhitespace()) {
            if (currentTag == "LatestStableMajor") {
                info->stable_major_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "LatestStableMinor") {
                info->stable_minor_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "LatestStableMicro") {
                info->stable_micro_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "DateLatestVersion") {
                info->date_latest_stable_version = atoi(xml.text().toString().toUtf8());
                if ((now = localtime(&info->date_latest_stable_version)) != NULL) {
                    strftime(info->date_latest_stable_version_str, sizeof(info->date_latest_stable_version_str), "%x", now);
                }
            } else if (currentTag == "LatestUnstableMajor") {
                info->unstable_major_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "LatestUnstableMinor") {
                info->unstable_minor_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "LatestUnstableMicro") {
                info->unstable_micro_version = atoi(xml.text().toString().toUtf8());
            } else if (currentTag == "DateLatestUnstable") {
                info->date_latest_unstable_version = atoi(xml.text().toString().toUtf8());
                if ((now = localtime(&info->date_latest_unstable_version)) != NULL) {
                    strftime(info->date_latest_unstable_version_str, sizeof(info->date_latest_unstable_version_str), "%x", now);
                }
            } else if (current_changelog != NULL && currentTag == "VersionMajor") {
                current_changelog->major_version = atoi(xml.text().toString().toUtf8());
            } else if (current_changelog != NULL && currentTag == "VersionMinor") {
                current_changelog->minor_version = atoi(xml.text().toString().toUtf8());
            } else if (current_changelog != NULL && currentTag == "VersionMicro") {
                current_changelog->micro_version = atoi(xml.text().toString().toUtf8());
            } else if (current_changelog != NULL && currentTag == "Entry-i18n") {
                if (!process_this_translated_entry) {
                    /* Different language, ignore */
                    continue;
                }

                current_changelog->entries.append(xml.text().toString());
                process_this_translated_entry = FALSE;
            } else if (current_changelog != NULL && !translated_entry_seen && currentTag == "Entry") {
                current_changelog->entries.append(xml.text().toString());
            }
        }
    }
    if (xml.error() && xml.error() != QXmlStreamReader::PrematureEndOfDocumentError) {
        qWarning() << "Error while parsing update information:" << xml.lineNumber() << ": " << xml.errorString() << endl << xml_data;
        return false;
    }

    return true;
}

#define HTML_INFO_VERSION \
"<TABLE WIDTH=\"50%%\" ALIGN=\"center\">" \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>" \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>" \
"</TABLE>"

#define HTML_INFO_UNSTABLE_VERSION \
"<TABLE WIDTH=\"50%%\" ALIGN=\"center\">" \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>" \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>" \
"<TR><TD>%s</TD><TD>%i.%i.%i</TD></TR>" \
"</TABLE>"

#define HTML_INFO_NO_UPDATE \
"<P ALIGN=\"center\">%s</P>"

#define HTML_INFO_SNAPSHOT \
"<P ALIGN=\"center\">%s<BR>" \
"%s %i (%s %i) %s %i</P>"

#define HTML_INFO_UNSTABLE_NO_UPDATE \
"<P ALIGN=\"center\">%s<BR/>%s</P>"

#define HTML_INFO_UPDATE \
"<H3 ALIGN=\"center\">%s</H3>" \
"<P ALIGN=\"center\">%s %i.%i.%i %s %s %s</P>"

#define HTML_INFO_CHANGELOG_BEGIN \
"<P><U><B>%s %i.%i.%i:</B></U>" \
"<UL>"

#define HTML_INFO_CHANGELOG_END \
"</UL></P>"

#define HTML_UNSTABLE_NOTICE \
"<P ALIGN=\"center\"><B>%s</B></P></CENTER>"

#define HTML_INFO_WEBSITE \
"<P ALIGN=\"center\">%s" \
"<A HREF=\"http://www.nntpgrab.nl\">http://www.nntpgrab.nl</A>" \
"</P>"

#define CHECK_VERSION_IS_OLDER(major,minor,micro) \
            (MAJOR_VERSION < (major) || \
            (MAJOR_VERSION == (major) && MINOR_VERSION < (minor)) || \
            (MAJOR_VERSION == (major) && MINOR_VERSION == (minor) && \
             MICRO_VERSION < (micro)))

#define CHECK_IS_VERSION(major,minor,micro) \
            (MAJOR_VERSION == (major) && MINOR_VERSION == (minor) && MICRO_VERSION == (micro))

void WidgetUpdateInfo::processAndShowUpdateInfo(QString xml_data)
{
    UpdateInfo info;
    QString str;
    QString res = "";

    if (!processUpdateInfo(xml_data, &info)) {
        return;
    }

    if (info.unstable_major_version >= 0) {
        res.append(str.sprintf(HTML_INFO_UNSTABLE_VERSION, tr("You are using version:").toUtf8().data(), MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION, tr("The latest available stable version is:").toUtf8().data(), info.stable_major_version, info.stable_minor_version, info.stable_micro_version, tr("The latest available unstable version is:").toUtf8().data(), info.unstable_major_version, info.unstable_minor_version, info.unstable_micro_version));
    } else {
        res.append(str.sprintf(HTML_INFO_VERSION, tr("You are using version:").toUtf8().data(), MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION, tr("The latest available version is:").toUtf8().data(), info.stable_major_version, info.stable_minor_version, info.stable_micro_version));
    }

    if (info.unstable_major_version < 0 && CHECK_IS_VERSION(info.stable_major_version, info.stable_minor_version, info.stable_micro_version)) {
        res.append(str.sprintf(HTML_INFO_NO_UPDATE, tr("You are using the latest version of NNTPGrab").toUtf8().data()));
        res.append(str.sprintf(HTML_INFO_WEBSITE, tr("For more information, go to our website ").toUtf8().data()));
    } else if (CHECK_IS_VERSION(info.unstable_major_version, info.unstable_minor_version, info.unstable_micro_version)) {
        res.append(str.sprintf(HTML_INFO_UNSTABLE_NO_UPDATE, tr("You are using the latest unstable version of NNTPGrab").toUtf8().data(), tr("If you find any bugs with this version, please report these").toUtf8().data()));
        res.append(str.sprintf(HTML_INFO_WEBSITE, tr("For more information, go to our website ").toUtf8().data()));
    } else if ((info.unstable_major_version < 0 && !(CHECK_VERSION_IS_OLDER(info.stable_major_version, info.stable_minor_version, info.stable_micro_version))) ||
               (info.unstable_major_version >=0 && !(CHECK_VERSION_IS_OLDER(info.unstable_major_version, info.unstable_minor_version, info.unstable_micro_version)))) {

        res.append(str.sprintf(HTML_INFO_SNAPSHOT, tr("You are using a snapshot version of NNTPGrab").toUtf8().data(), tr("This is subversion revision").toUtf8().data(), __BUILD_SVN_REV, tr("build").toUtf8().data(), __BUILD_NUMBER, tr("which was built on ").toUtf8().data(), __BUILD_DATE));
        res.append(str.sprintf(HTML_INFO_WEBSITE, tr("For more information, go to our website ").toUtf8().data()));
    } else {
        if (info.unstable_major_version >= 0) {
            res.append(str.sprintf(HTML_INFO_UPDATE, tr("A new unstable version of NNTPGrab is available!").toUtf8().data(), tr("Version").toUtf8().data(), info.unstable_major_version, info.unstable_minor_version, info.unstable_micro_version, tr("is released on").toUtf8().data(), info.date_latest_unstable_version_str, tr("and contains the following changes :").toUtf8().data()));
        } else {
            res.append(str.sprintf(HTML_INFO_UPDATE, tr("A new version of NNTPGrab is available!").toUtf8().data(), tr("Version").toUtf8().data(), info.stable_major_version, info.stable_minor_version, info.stable_micro_version, tr("is released on").toUtf8().data(), info.date_latest_stable_version_str, tr("and contains the following changes :").toUtf8().data()));
        }

        for (int i = 0; i < info.changelogs.length(); i++) {
            ChangeLog *changelog = info.changelogs.at(i);

            if (!CHECK_VERSION_IS_OLDER(changelog->major_version, changelog->minor_version, changelog->micro_version)) {
                continue;
            }

            res.append(str.sprintf(HTML_INFO_CHANGELOG_BEGIN, tr("Changes in version").toUtf8().data(), changelog->major_version, changelog->minor_version, changelog->micro_version));

            for (int j = 0; j < changelog->entries.length(); j++) {
                res.append(str.sprintf("<LI>%s", changelog->entries.at(j).toUtf8().data()));
            }

            res.append(QString(HTML_INFO_CHANGELOG_END));
        }

        if (info.unstable_major_version >= 0) {
            res.append(str.sprintf(HTML_UNSTABLE_NOTICE, tr("Please note that the unstable version may still contain bugs, so only use the unstable version when you want to help out testing").toUtf8().data()));
        }

        res.append(str.sprintf(HTML_INFO_WEBSITE, tr("To retrieve the latest version and for more information, go to our website ").toUtf8().data()));
    }

    ui->webViewUpdateInfo->append(res);
}
