/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NNTPCONNECTION_BACKEND_IFACE_H_
#define _NNTPCONNECTION_BACKEND_IFACE_H_

#include <sys/time.h>
#include <glib-object.h>

#ifdef WIN32
#include <winsock2.h>
#else
#include <netdb.h>
#endif

#include "configuration.h"
#include "collections.h"

G_BEGIN_DECLS

typedef enum {
    NNTP_JOB_TYPE_NONE,
    NNTP_JOB_TYPE_ARTICLE,
    NNTP_JOB_TYPE_XOVER
} NNTPJobType;

typedef enum {
    NNTP_CONNECTION_ERROR_NONE,
    NNTP_CONNECTION_ERROR_HOST_NOT_FOUND,
    NNTP_CONNECTION_ERROR_SSL_INITIALISE,
    NNTP_CONNECTION_ERROR_CONNECTION_REFUSED,
    NNTP_CONNECTION_ERROR_CONNECTION_TIMEOUT,
    NNTP_CONNECTION_ERROR_READ,
    NNTP_CONNECTION_ERROR_WRITE,
    NNTP_CONNECTION_ERROR_INVALID_MSG,
    NNTP_CONNECTION_ERROR_INVALID_ARGUMENT,
    NNTP_CONNECTION_ERROR_LOGIN_FAILURE,
    NNTP_CONNECTION_ERROR_TOO_MANY_CONNECTIONS,
    NNTP_CONNECTION_ERROR_PART_NOT_AVAILABLE,
    NNTP_CONNECTION_ERROR_NO_SUCH_GROUP,
    NNTP_CONNECTION_ERROR_UNABLE_TO_SAVE_PART,
} NNTPConnectionErrCode;

#define NNTP_TYPE_CONNECTION                (nntpconnection_get_type ())
#define NNTP_CONNECTION(obj)                (G_TYPE_CHECK_INSTANCE_CAST ((obj), NNTP_TYPE_CONNECTION, NntpConnection))
#define NNTP_IS_CONNECTION(obj)             (G_TYPE_CHECK_INSTANCE_TYPE ((obj), NNTP_TYPE_CONNECTION))

#define NNTPCONNECTION_TYPE_BACKEND                (nntpconnection_backend_get_type ())
#define NNTPCONNECTION_BACKEND(obj)                (G_TYPE_CHECK_INSTANCE_CAST ((obj), NNTPCONNECTION_TYPE_BACKEND, NntpconnectionBackend))
#define NNTPCONNECTION_IS_BACKEND(obj)             (G_TYPE_CHECK_INSTANCE_TYPE ((obj), NNTPCONNECTION_TYPE_BACKEND))
#define NNTPCONNECTION_BACKEND_GET_INTERFACE(inst) (G_TYPE_INSTANCE_GET_INTERFACE ((inst), NNTPCONNECTION_TYPE_BACKEND, NntpconnectionBackendIface))

typedef struct _NntpConnection NntpConnection;
typedef struct _NntpConnectionClass NntpConnectionClass;

typedef struct _NntpconnectionBackend NntpconnectionBackend;    /* dummy interface */
typedef struct _NntpconnectionBackendIface NntpconnectionBackendIface;

struct _NntpConnection {
    GObject parent;

    int connection_id;
    int server_id;
    NGConfigServer server_info;

    GPollFD poll_fd;

    void (*active_method) (NntpConnection *conn);
    int max_bandwidth;

    gboolean is_idle;
    time_t idle_start_stamp;
    time_t last_checked_stamp;
    time_t last_disconnect_stamp;
    time_t last_activity_stamp;

    NNTPJobType job_type;

    NNTPCollection *collection;
    NNTPFile *file;
    NNTPPart *part;

    char newsgroup[256];
    gint64 xover_start_range;
    gint64 xover_end_range;

    char article_filename[1024];
    int article_fd;
    int article_bytes_downloaded;
    struct timeval last_article_progress_announce;
    char article_write_buffer[50000];              /* Flush the received article data every 50.000 bytes, GObject doesn't support larger objects */
    int article_write_buffer_length;
};

struct _NntpconnectionBackendIface {
    GTypeInterface parent;

    NNTPConnectionErrCode (*connect) (NntpConnection *self, const char *hostname, int port, char **errmsg);
    void (*disconnect) (NntpConnection *self);

    gboolean (*prepare_ssl) (NntpConnection *conn, char **errmsg);

    gboolean (*ensure_data_is_ready) (NntpConnection *conn, int timeout);

    gboolean (*read_msg) (NntpConnection *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready, gboolean *try_again, char **errmsg);
    gboolean (*send_msg) (NntpConnection *conn, void *data, int length, char **errmsg);

    gboolean (*get_has_ssl_support) (NntpConnection *conn);
};

GType nntpconnection_get_type (void);
GType nntpconnection_backend_get_type (void);

void nntpconnection_backend_init_traffic_monitor(void);
void nntpconnection_backend_update_traffic_monitor(int bytes_read);
void nntpconnection_backend_perform_shaping(int max_bandwidth, int bytes_read);

/* Wrapper functions */
NNTPConnectionErrCode nntpconnection_backend_connect(NntpConnection *conn, const char *hostname, int port, char **errmsg);
void nntpconnection_backend_disconnect(NntpConnection *conn);
gboolean nntpconnection_backend_prepare_ssl(NntpConnection *conn);

/* Public functions */
gboolean nntpconnection_get_has_ssl_support(NntpConnection *conn);
NNTPConnectionErrCode nntpconnection_connect_to_server(NntpConnection *conn, char **errmsg);
void nntpconnection_disconnect_from_server(NntpConnection *conn, NNTPDisconnectType disconnect_type, const char *reason, const char *cause_file, int cause_lineno);

gboolean nntpconnection_ensure_data_is_ready(NntpConnection *conn, int timeout);
gboolean nntpconnection_read_msg(NntpConnection *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready);
gboolean nntpconnection_send_msg(NntpConnection *conn, void *data, int length);
gboolean nntpconnection_get_has_ssl_support(NntpConnection *conn);

void nntpconnection_process_proxy_connect_command(NntpConnection *conn);

G_END_DECLS

#endif /* _NNTPCONNECTION_BACKEND_IFACE_H_ */
